# Author::    Jens Bachmann  (mailto:jens.bachmann@unibw.de)
# Author::		Daniel Borkowitz (mailto:daniel.borkowitz@unibw.de)
# Author::		Stefan Mayer (mailto:stefan.mayer@unibw.de)
# Author::		Roman Krhling (mailto:roman.kraehling@unibw.de)
# Copyright:: Copyright (c) 2006 The Heimschlfers, UniBw Mnchen

# Klasse Menge dient zur Modellierung von Mengen, hierbei werden die Elemente der Menge
# in einem Ruby-Array (@array) gehalten
class Menge

# Erschafft ein Mengenobjekt aus einem Arrayparameter, entfernt 
# Doubletten und sortiert die Menge
	def initialize(array)
		@array = array
		@array.sort!
		@array.uniq!
	end
	
# Konvertiert die Menge in String 
# Bsp.: [1,2,3,4,5] -> "{ 1,2,3,4,5 }"
	def to_s
		return "{}" if @array == nil
		n=0
		str = "{ " + @array[0].to_s
		while n<=@array.length
			n+=1
			str += ',' + @array[n].to_s if @array[n]!=nil
		end
		return str + " }"
	end
	
	def [](i)
		@array[i]
	end
	
# Fgt ein Element an die Menge an und sortiert die Menge und entfernt Doubletten
	def add(elem)
		(self.enthaelt?(elem))?():@array+=[elem]
		@array.uniq!
		@array.sort!
		return self
	end
	
# Prft ob die Menge das Element enthlt
	def enthaelt?(x)
		@array.include?(x)
	end

	
# Getter fr @array	
	def array()
		@array
	end

# Liefert die Anzahl der Elemente der Menge
	def length()
		@array.length()
	end

# Vereinigung zweier Mengen, dabei wird berprft ob die Menge (Parameter) nil ist	
	def +(menge)
		if menge!=nil then 
				temp = Menge.new([])
				each{|i| temp.add(i)}
				menge.each{|i| temp.add(i)}
				temp 
		else return Menge.new(@array) end
	end

# Differenz zweier Mengen, dabei wird berprft ob die Menge (Parameter) nil ist	
	def -(menge)
		if menge!=nil then Menge.new(@array - menge.array()) else return Menge.new(@array) end
	end
	
	def <=>(another)
		@array <=> another.array
	end
	

# Prft Gleichheit zweier Mengen, dabei wird berprft ob die Menge (Parameter) nil ist	
	def ==(menge)
		if menge!=nil then (@array==menge.array()) else return false end
	end
	
# prft ob Menge leer ist
	def ist_leer?
	  @array.isEmpty?
	end

# Ist Iterator ber die Elemente der Menge
	def each &b
		@array.each &b
	end

# Liefert eine wahre Kopie der Menge	
	def copy()
		temp = Menge.new([])
		each{|i| temp.add(i)}
		temp
	end

# Speichert das Objekt Menge auf der Harddisk mit Dateiname als Parameter mittels Marshall	
	def speichern(dateiname)
		File.open(dateiname, "w") do |f|
			Marshal.dump(self,f)
		end
	end

# Ldt das Objekt Menge von der Harddisk mit Dateiname als Parameter mittels Marshall	
	def laden(dateiname)
		File.open(dateiname, "r") do |f|
			@array+=Marshal.load(f).array()
		end
	end
end


#######

# Klasse Relation dient zur Modellierung von Relationen, hierbei wird die Basismenge im 
# Attribut @basis als Menge gehalten. Zur Basis gibt es eine Vorgnger- und eine 
# Nachfolgeradjazenzliste (@vor und @nach). Diese sind als Ruby-Hashs von Mengen abgelegt,
# der Schlssel ist jeweils das Element aus der Basis.
class Relation

# Konstruktor der Klasse Relation erhlt ein Array von Paaren (ein Paar ist ein Array mit zwei Elementen).
# Fr @vor und @nach werden neue Hashs angelegt. Die Basis wird anhand des bergegebenen Parameters (rel) berechnet.
# Die Vorgnger und Nachfolger werden nun anhand der Basis und rel berechnet.
	def initialize(rel)
		@vor = Hash.new()
		@nach = Hash.new()
		@basis=Menge.new([]) 
		if rel!=[[]]
		@basis=Menge.new([])
		rel.each{|i| @basis.add(i[0]);@basis.add(i[1])}
	  each {|i| @vor[i]=berechneVorgaenger(i,rel)}
	  each {|i| @nach[i]=berechneNachfolger(i,rel)}
		end
	end

private
# Berechnet die Vorgnger eines Elements und gegebener Relation.
	def berechneVorgaenger(elem,rel)
		temp= Menge.new([])
		rel.each{|i| temp.add(i[0]) if i[1]==elem}
		return temp
	end
	
# Berechnet die Nachfolger eines Elements und gegebener Relation.
	def berechneNachfolger(elem,rel)
		temp= Menge.new([])
		rel.each{|i| temp.add(i[1]) if i[0]==elem}
		return temp
	end
	
public
# Getter fr die Vorgngeradjazenzliste.
	def vor 
		@vor
	end

# Setter fr die Vorgngeradjazenzliste.
	def vor!(vor)
		@vor=vor
	end
	
# Getter fr die Nachfolgeradjazenzliste.
	def nach
		@nach
	end

# Setter fr die Nachfolgeradjazenzliste.
	def nach!(nach)
		@nach=nach
	end
	
# Getter fr die Basismenge.
	def basis
		@basis
	end
	
# Setter fr die Basismenge.
	def basis!(basis)
		@basis = basis
	end

# Konvertiert die Relation in einen String.
# 
# Bsp.: [[1,2],[2,3],[3,4],[2,5]]
#
# "Basismenge:
#    { 1,2,3,4,5 }
# Adjazenzliste (Vorgaenger):
#   5 { 2 }
#		1 {  }
#   2 { 1 }
#   3 { 2 }
#   4 { 3 }
# Adjazenzliste (Nachfolger):
#   5 {  }
#   1 { 2 }
#   2 { 3,5 }
#   3 { 4 }
#   4 {  }"
	def to_s
			return "{}" if @basis == nil
			str = "Basismenge: \n   " + @basis.to_s + "\n" +
			"Adjazenzliste (Vorgaenger): \n   "
			@vor.each_key{|i| str+= "\t"+i.to_s+" " + (@vor[i].to_s) + "\n"}
			str+= "\n"
			str+="Adjazenzliste (Nachfolger):\n   "
      @nach.each_key{|i| str+= "\t"+i.to_s+" " + (@nach[i].to_s) + "\n"}
			return str
	end

# Speichert das Objekt Relation auf der Harddisk mit Dateiname als Parameter mittels Marshall	
	def speichern(dateiname)
		File.open(dateiname, "w") do |f|
			Marshal.dump(@basis,f)
			Marshal.dump(@vor,f)
			Marshal.dump(@nach,f)
		end
	end

# Ldt das Objekt Relation von der Harddisk mit Dateiname als Parameter mittels Marshall	
	def laden(dateiname)
		File.open(dateiname, "r") do |f|
			@basis=Marshal.load(f)
			@vor=Marshal.load(f)
			@nach=Marshal.load(f)
		end
	end
	
# Fgt ein Tupel (x,y) zur Relation hinzu. Hierbei wird die Relation verndert.
	def addPaar!(x,y) 
					@basis.add(x)
					@basis.add(y)
          @nach[x] = Menge.new([y]) + @nach[x] 
          @nach[y] = Menge.new([]) if @nach[y]==nil 
          @vor[y] = Menge.new([x]) + @vor[y] 
          @vor[x] = Menge.new([]) if @vor[x]==nil 
					return self
     end

# Prft ob die Relation das Tupel (x,y) enthlt.
	def enthaeltPaar?(x,y)
		return (@nach[x].enthaelt?(y) && @vor[y].enthaelt?(x))
	end

# Liefert die Menge der Vorgnger fr das Element x aus der Basismenge.
	def vorgaenger(x)
		@vor[x]!=nil ? @vor[x] : Menge.new([])
	end

# Liefert die Menge der Nachfolger fr das Element x aus der Basismenge.
	def nachfolger(x)
		@nach[x]!=nil ? @nach[x] : Menge.new([])
	end

# Prft Gleichheit zweier Relationen, dabei wird berprft ob die Relation (Parameter) nil ist.	
	def ==(rel)
		rel!=nil ? ((@basis==rel.basis) && (@vor==rel.vor) && (@nach==rel.nach)): false
	end
	
# Iterator ber die Elemente der Basismenge der Relation.
	def each &b
		@basis.each &b
	end

# Vereinigung zweier Relationen, dabei werden die Basen zusammengefasst und danach ber die entstandene
# Basis iteriert. Dabei werden die Vorgnger und Nachfolger aus beiden Relationen neu berechnet.
	def vereinigt(rel)
    erg = Relation.new([[]])
    erg.basis!(@basis + rel.basis)
	  erg.each{|i| erg.nach[i] = (@nach[i]==nil ? Menge.new([]) : @nach[i]) + rel.nach[i] }
    erg.each{|i| erg.vor[i]  = (@vor[i]==nil ? Menge.new([]) : @vor[i])  + rel.vor[i] }
    return erg
  end 
	
# Komposition zweier Relationen, dabei werden die Basen zusammengefasst und danach ber die entstandene
# Basis iteriert. Dabei werden die Vorgnger und Nachfolger aus beiden Relationen neu berechnet.
	def komp(rel) 
          erg=Relation.new([[]]) 
          erg.basis!(@basis+rel.basis) 
               erg.each{|i| erg.nach[i]=Menge.new([]); erg.vor[i]=Menge.new([])} 
               erg.each{|i| @nach[i].each{|j| erg.nach[i]+=rel.nachfolger(j)}if @nach[i]!=nil} 
               erg.each{|i| rel.vor[i].each{|j| erg.vor[i]+=self.vorgaenger(j)}if rel.vor[i]!=nil} 
               return erg 
     end


# Liefert eine wahre Kopie der Relation und ihrer Attribute.
	def copy()
		temp = Relation.new([[]])
		@basis.each{|i| temp.basis.add(i)}
		@vor.each_key{|i| temp.vor[i]=@vor[i].copy}
		@nach.each_key{|i| temp.nach[i]=@nach[i].copy}
		return temp
	end

# Berechnet die reflexive Hlle der Relation.
	def reflexiv
		temp = copy
		temp.each{|i| temp.addPaar!(i,i)}
		return temp
	end

# Berechnet die transitive Hlle der Relation mittels Warshall-Algorithmus.
	def transitiv
		temp = copy
		temp.basis.length().times do
		temp.each{|i| temp.nachfolger(i).each{|j| temp.nach[i]+=temp.nachfolger(j)}}
		temp.each{|i| temp.nachfolger(i).each{|j| temp.vor[j]+=Menge.new([i])}}
		end
		return temp
	end
	# Entfernt ein Tupel (x,y) aus der Relation. Hierbei wird die Relation verndert.
	def entfernePaar(x,y)
			if enthaeltPaar?(x,y) 
	        @nach[x] =  @nach[x] - Menge.new([y])
          @vor[y] =   @vor[y] - Menge.new([x])
					return self
			end
	end
	# Entfernt "Nonterminal -> Nonterminal" (Konvention Grobuchstabe) aus der Relation
	def entferneNonterminale
	   temp = self.copy()
	   temp.each{|i| temp.nach[i].each{|j| temp.entfernePaar(i,j) if (('A'<=j and j<='Z')or(j=="@"))}}
		 return temp
	end
	
	def inv
		temp = self.copy
		ttemp = temp.nach
		temp.nach!(temp.vor)
		temp.vor!(ttemp)
		return temp
	end
	
end

#Testflle der Klasse Menge
require "test/unit"
class MengeToTest < Test::Unit::TestCase

	def setup
		@menge1 = Menge.new([5,4,3,2,1])
		@menge2 = Menge.new([9,8,6,7,5])
		@menge3 = Menge.new([1,2,3,4,5,6,7,8,9])
		@menge4 = Menge.new([1,2,3,4])
	end
	def test_plus
		assert_equal(@menge1 + @menge2, @menge3)
	end

	def test_minus
		assert_equal(@menge3 - @menge2, @menge4)
	end
	
	def test_gleich
		assert(@menge1 == @menge1)
		assert(@menge3 == @menge3)
		assert(!(@menge1 == @menge2))
		assert(!(@menge1 == @menge4))
	end

	def test_add
		assert_equal(@menge4.add(5), @menge1)
		assert_equal(@menge1.add(5), @menge1)
	end

	def test_copy
		temp = @menge4.copy
		assert_not_same(temp, @menge4)
		assert_equal(temp, @menge4)
	end

	def test_each
		temp = Menge.new([])
		@menge1.each{|i| temp.add(i)}
		assert_equal(temp, @menge1)
	end

	def test_enthaelt?
		@menge1.each{|i| assert(@menge1.enthaelt?(i))}
		assert(!(@menge2.enthaelt?(1)))
	end
	
	def test_speichernladen
		temp = Menge.new([])
		@menge1.speichern("testfile.txt")
		temp.laden("testfile.txt")
		assert_equal(@menge1,temp)
	end
	
	def test_length
		assert_equal(@menge1.length(),5)
	end
	
end

#Testflle der Klasse Relation
class RelationToTest < Test::Unit::TestCase

	def setup
		@rel1 = Relation.new([[1,1],[1,2],[2,3],[2,1],[4,4]])
		@rel2 = Relation.new([[1,2],[2,3],[3,4],[2,5]])
		@rel3 = Relation.new([[1,2],[2,3],[3,4],[2,5],[5,5]])
		@rel4 = Relation.new([[1,1],[2,2],[3,3],[4,4]])
		@rel5 = Relation.new([[1,2],[2,3],[4,5],[1,3]])
		@rel6 = Relation.new([[2,1],[3,2],[5,4],[3,1]])
	end

	def test_gleich
		assert(@rel1 == @rel1)
		assert(@rel2 == @rel2)
		assert(!(@rel1 == @rel2))
	end

	def test_addPaar
		assert_equal(@rel2.addPaar!(5,5), @rel3)
		assert_equal(@rel2.addPaar!(2,5), @rel2)
	end

	def test_copy
		temp = @rel2.copy
		assert_not_same(temp, @rel2)
		assert_equal(temp, @rel2)
	end

	def test_each
		temp = Relation.new([])
		@rel1.each{|i| temp.addPaar!(i,i)}
		assert_equal(temp, @rel4)
	end

	def test_speichernladen
		temp = Relation.new([])
		@rel1.speichern("testfile.txt")
		temp.laden("testfile.txt")
		assert_equal(@rel1,temp)
	end
	
	def test_enthaeltPaar?
		@rel4.each{|i| assert(@rel4.enthaeltPaar?(i,i))}
		assert(!(@rel1.enthaeltPaar?(2,2)))
	end
	
	def test_berechnevorgaenger
		temp = Menge.new([1,2])
		assert_equal(@rel1.vorgaenger(1), temp)
	end

	def test_berechnenachfolger
		temp = Menge.new([1,3])
		assert_equal(@rel1.nachfolger(2), temp)
	end
	
	def test_komp
		temp = Relation.new([[1,2],[1,3],[1,5],[2,4],[2,2]])
		assert_equal(temp, @rel1.komp(@rel2))
	end
	
	def test_vereinigt
		temp = Relation.new([[1,1],[1,2],[2,3],[2,1],[4,4],[1,2],[2,3],[3,4],[2,5]])
		temp= Relation.new([[]])
		assert_equal(@rel1, @rel1.vereinigt(temp))
	end
	
	def test_reflexiv
		temp = @rel2.reflexiv
		@rel2.each{|i| assert(temp.enthaeltPaar?(i,i))}
	end
	
	def test_transitiv
		temp = Relation.new([[1,2],[1,3],[1,4],[1,5],[2,3],[2,4],[2,5],[3,4]])
		assert_equal(@rel2.transitiv, temp)
	end
	
	def test_inv
		assert_equal(@rel5.inv, @rel6)
	end
end



###############################################


=begin
puts eins = Menge.new([1,2,3])
puts eins+(Menge.new([4]))
puts eins-(Menge.new([4]))
puts
puts eins.enthaelt?(2)
puts eins.enthaelt?('a')
puts
puts eins.length
puts
eins.speichern("testspeichern")
zwei = Menge.new([])
zwei.laden("testspeichern")
puts zwei
puts eins==zwei
=end

=begin
puts rel1= Relation.new([[1,2],[2,3],[3,4]])
gets
puts rel1.addPaar!(4,5)
gets
puts rel1.entfernePaar(4,5)
gets
puts rel1.enthaeltPaar?(1,2)
rel1.speichern("speicherntest")
rel2 = Relation.new([])
rel2.laden("speicherntest")
puts rel2
puts rel1==rel2
=end


=begin
rel1 = Relation.new([[1,2],[2,3],[3,4]])
rel2 = Relation.new([[2,3],[4,5]])
temp = rel1.vereinigt(rel2)
puts temp
gets
temp = rel1.komp(rel2)
puts temp
gets
temp = rel1.reflexiv
puts temp
gets
temp = rel1.transitiv
puts temp
gets
temp = rel1.reflexiv.transitiv
puts temp
=end


=begin
endet_mit=Relation.new([['E','T'],['T','F'],['F','('],['F',')'],['F','id']])
puts ((endet_mit.transitiv()).entferneNonterminale())
=end



####################################################
=begin
puts ">"
inp = gets.chop
while inp != ""
  puts "#{inp} = #{eval(inp)}"
  inp = gets.chop
end
=end
